# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Implementations of draft-ietf-core-resource-directory

Currently, this only contains the :mod:`client` module. There is an
implementation of the server side contained in aiocoap, but it is currently
accessible only as :mod:`command line application <aiocoap.cli.rd>`.
"""
