# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Container module for command line utilities bundled with aiocoap.

These modules are not considered to be a part of the aioCoAP API, and are thus
subject to change even when the project reaches a stable version number. If you
want to use any of that infrastructure, please file a feature request for
stabilization in the project's issue tracker.

The tools themselves are documented in :doc:`/tools`.
"""
