# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Client components for draft-ietf-core-resource-directory

Currently, this only contains the :mod:`register` module, which helps utilizing
the registration interface of a resource directory.
"""
