# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Container module, see submodules:

* :mod:`.client` -- using CoAP via a proxy server
* :mod:`.server` -- running a proxy server
"""
